class InvalidZygosity(Exception):
    def __init__(self, chromosome, position, zygosity):
        self.chromosome = chromosome
        self.position = position
        self.zygosity = zygosity

    def __str__(self):
        return f'Got invalid zygosity in {self.chromosome} at position {self.position}: {self.zygosity}'


class InvalidVariantLine(Exception):
    def __init__(self, line):
        self.line = line

    def __str__(self):
        return f'Variant line from vcf is invalid: {self.line}'


import re

# The following regular expressions were taken from PyVCF
# https://github.com/jamescasbon/PyVCF/blob/476169cd457ba0caa6b998b301a4d91e975251d9/vcf/parser.py#L84-L111

INFO_PATTERN = re.compile(r'''\#\#INFO=<
            ID=(?P<id>[^,]+),\s*
            Number=(?P<number>-?\d+|\.|[AGR]),\s*
            Type=(?P<type>Integer|Float|Flag|Character|String),\s*
            Description="(?P<desc>[^"]*)"
            (?:,\s*Source="(?P<source>[^"]*)")?
            (?:,\s*Version="?(?P<version>[^"]*)"?)?
            >''', re.VERBOSE)

FILTER_PATTERN = re.compile(r'''\#\#FILTER=<
            ID=(?P<id>[^,]+),\s*
            Description="(?P<desc>[^"]*)"
            >''', re.VERBOSE)

ALT_PATTERN = re.compile(r'''\#\#ALT=<
            ID=(?P<id>[^,]+),\s*
            Description="(?P<desc>[^"]*)"
            >''', re.VERBOSE)

FORMAT_PATTERN = re.compile(r'''\#\#FORMAT=<
            ID=(?P<id>.+),\s*
            Number=(?P<number>-?\d+|\.|[AGR]),\s*
            Type=(?P<type>.+),\s*
            Description="(?P<desc>.*)"
            >''', re.VERBOSE)

CONTIG_PATTERN = re.compile(r'''\#\#contig=<
            ID=(?P<id>[^>,]+)
            (,.*length=(?P<length>-?\d+))?
            .*
            >''', re.VERBOSE)

META_PATTERN = re.compile(r'''##(?P<id>.+?)=(?P<val>.+)''')

GT_PATTERN = re.compile('([0-9]+|\.)[/|]([0-9]+|\.)')



from enum import Enum


class Zygosity(str, Enum):
    homozygote = 'HOM'
    heterozygote = 'HET'
    reference = 'REF'
    no_coverage = 'NO_COVERAGE'

    def __str__(self):
        return self.value



import hashlib
import re
from cmath import nan
from typing import AnyStr
from typing import Iterable
from typing import TextIO

from vcfio.utils.consts import VALID_CHROMOSOMES
from vcfio.utils.enums import Zygosity
from vcfio.utils.exceptions import InvalidVariantLine
from vcfio.utils.exceptions import InvalidZygosity
from vcfio.utils.str_utils import calculate_zygosity
from vcfio.variant.variant_properties import VariantProperties


class Variant(VariantProperties):
    # A class with implementation of methods

    def is_sv(self) -> bool:
        """ Return whether the variant is a structural variant """
        return self.info.get('SVTYPE') is not None

    def stringify_info(self):
        fields = []
        for key, val in self.info.items():
            if val is nan:
                res_string = key
            else:
                if isinstance(val, list):
                    val = ",".join(map(str, val))
                res_string = f"{key}={val}"
            fields.append(res_string)
        return ';'.join(fields).replace(" ", "_")

    def stringify_samples(self):
        def stringify_sample(value):
            """This method converts the value to string depending on the key"""
            if isinstance(value, list):
                converted_list = [str(element) if element is not None else '.' for element in value]
                result = ",".join(converted_list)
            else:
                result = str(value) if value is not None else '.'
            return result

        return [':'.join([stringify_sample(sample.get(format_field)) for format_field in self.sample_format])
                for sample in self.samples.values()]

    def to_vcf_line(self, with_samples: bool = True):
        return '\t'.join(map(str, [
            self.chromosome,
            self.position,
            self.vcf_id or '.',
            self.ref,
            ','.join(self.alt),
            self.quality or '.',
            ';'.join(self.vcf_filter) or '.',
            self.stringify_info() or '.',
        ] + ([':'.join(self.sample_format), *self.stringify_samples()] if with_samples else [])
                             )).strip()

    def get_zygosity(self, sample_name) -> Zygosity:
        gt = self.samples[sample_name].get('GT', './.')
        try:
            return calculate_zygosity(gt)
        except IndexError:
            raise InvalidZygosity(self.chromosome, self.position, gt)

    def is_sample_called(self, sample_name):
        return '.' not in self.samples[sample_name].get('GT', '.')

    def is_chromosome_valid(self):
        return self.chromosome in VALID_CHROMOSOMES

    @classmethod
    def from_variant_line(cls, variant_line: TextIO, sample_names: Iterable[AnyStr] = (),
                          default_sample_format: Iterable[AnyStr] = ()) -> 'Variant':
        """
        Create a Variant instance from a raw line
        Example:
            Input - "chr3    2956    .       G       GACACACAC       100     .       AC=1;AN=2;DP4=1,0,6,0;DP=7;IDV=6;IMF=0.857143;INDEL;MQ0F=0;MQ=51;SGB=-0.616816;VDB=0.041536;multiallele.gid=1   GT:PL:AD        0/1:111,0,113:0,4"
            Output - Variant(chromosome=chr3,position=2956,ref=G,alt=['GACACACAC'])
        """
        variant_line_values = re.split('\t| +', variant_line.strip('\n'))

        if len(variant_line_values) < 8:
            raise InvalidVariantLine(variant_line)

        chromosome, position, vcf_id, ref, alt, quality, vcf_filter, info = variant_line_values[:8]
        sample_format, *samples = variant_line_values[8:] or ('',)

        new_variant = cls(
            chromosome=chromosome,
            position=position,
            vcf_id=vcf_id,
            ref=ref,
            alt=alt,
            quality=quality,
            vcf_filter=vcf_filter,
            info=info,
            sample_format=sample_format or default_sample_format,
            sample_names=sample_names,
            samples=samples
        )

        return new_variant


from __future__ import annotations

import functools
import logging
from pathlib import Path

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import AnyStr
    from typing import Iterator
    from typing import List
    from typing import Union


from vcfio.utils.file_utils import open_file
from vcfio.utils.regex_patterns import ALT_PATTERN
from vcfio.utils.regex_patterns import CONTIG_PATTERN
from vcfio.utils.regex_patterns import FILTER_PATTERN
from vcfio.utils.regex_patterns import FORMAT_PATTERN
from vcfio.utils.regex_patterns import INFO_PATTERN
from vcfio.utils.regex_patterns import META_PATTERN
from vcfio.utils.str_utils import standardize_chromosome
from vcfio.utils.str_utils import to_number
from vcfio.variant.variant import Variant


class VcfReader:
    def __init__(self, input_file: Union[Path, AnyStr]):
        self.input_file = Path(input_file)
        self.contigs = {}
        self.infos = {}
        self.filters = {}
        self.alts = {}
        self.formats = {}
        self.metadata = {}
        self._default_sample_format = ''
        self._file_descriptor = open_file(self.input_file)
        self._raw_variant_iterator = self._file_descriptor
        _ = self.headers

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False

    def close(self):
        self._file_descriptor.close()

    def __iter__(self) -> Iterator[Variant]:
        return self

    def __next__(self) -> Variant:
        """
        Iterate variants and yield Variant instances
        If sample format is not given in the raw line then use the FIRST format in the file
        """
        line = next(self._raw_variant_iterator)
        if not line or line.isspace():
            raise StopIteration

        variant = Variant.from_variant_line(line, self.sample_names, self._default_sample_format)
        self._default_sample_format = self._default_sample_format or variant.sample_format
        return variant

    @property
    @functools.lru_cache(maxsize=None)
    def headers(self) -> List[AnyStr]:
        """
        Read first lines of self._reader until last header (#CHROM  POS ID  REF...)
        Match regex patterns of header types and initialize them to instance attributes
        Return raw header lines
        """
        raw_header_lines = []
        for raw_line in self._file_descriptor:
            line = raw_line.strip()
            self._match_header_pattern(line)
            raw_header_lines.append(line)
            if line[1:].lower().startswith("chrom"):
                return raw_header_lines
        return []

    def _match_header_pattern(self, raw_header_line: AnyStr):
        attribute_patterns = (
            (self.infos, INFO_PATTERN),
            (self.contigs, CONTIG_PATTERN),
            (self.filters, FILTER_PATTERN),
            (self.alts, ALT_PATTERN),
            (self.formats, FORMAT_PATTERN),
            (self.metadata, META_PATTERN)
        )
        for attribute, pattern in attribute_patterns:
            match = pattern.match(raw_header_line)
            if match:
                attribute[match.group('id')] = to_number(match.groupdict(), default=match.groupdict())

    @property
    def variant_fields(self) -> List[AnyStr]:
        # CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	proband
        return self.headers[-1].strip('#').strip('\n').split()

    @property
    def sample_names(self) -> List[AnyStr]:
        # CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	---> [proband, mother, father] <---
        return self.variant_fields[9:]

    def fetch(self, chromosome: AnyStr, start: int = 1, end: int = None) -> Iterator[Variant]:
        """
        ** This function fetches using a 1-based system **
        If there is a tabix file - fetch using pysam
        Else - open another reader of the same path and iterate the selected variant
        """
        try:
            yield from self._fetch_by_index(chromosome, end, start)
        except OSError:  # .tbi file not found
            logging.warning(
                f"Tab-index file not found ({self.input_file.as_posix() + '.tbi'}) - using regular iteration.")
            yield from self._fetch_by_iteration(chromosome, end, start)
        except ModuleNotFoundError:  # if pysam is not installed
            logging.warning("Pysam module not found - using regular iteration.")
            yield from self._fetch_by_iteration(chromosome, end, start)

    def _fetch_by_index(self, chromosome, end, start):
        from pysam import TabixFile
        with TabixFile(filename=self.input_file.as_posix(), encoding='utf-8') as tabix_file:
            for line in tabix_file.fetch(standardize_chromosome(chromosome), start - 1, end - 1 if end else None):
                yield Variant.from_variant_line(
                    line,
                    sample_names=self.sample_names,
                    default_sample_format=self._default_sample_format
                )

    def _fetch_by_iteration(self, chromosome, end, start):
        with VcfReader(self.input_file) as sub_reader:
            for variant in sub_reader:
                if variant.chromosome == chromosome and variant.position in range(start, end):
                    yield variant
                if variant.position == end:
                    break

    def __repr__(self):
        return f"VcfReader('{self.input_file.absolute()}')"



from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import TextIO

def get_open_method(path: Path, mode, encoding='utf-8'):
    from gzip import open as gzip_open
    try:
        from Bio.bgzf import open as bgzip_open
    except ModuleNotFoundError:
        logging.warning("Biopython module not found - using gzip. It is recommended to pip install biopython.")
        bgzip_open = gzip_open

    if 'b' in mode:
        encoding = None

    if 'r' in mode:
        with path.open(mode='rb') as fin:
            header = fin.read(4)
        if header == b'\x1f\x8b\x08\x08':
            return gzip_open(path, mode, encoding=encoding)
        if header == b'\x1f\x8b\x08\x04':
            try:
                return bgzip_open(path, mode, encoding=encoding)
            except TypeError:
                return bgzip_open(path, mode)
    elif '.gz' in path.suffixes or '.bgz' in path.suffixes:
        try:
            return bgzip_open(path, mode, encoding=encoding)
        except TypeError:
            return bgzip_open(path, mode)

    return open(path, mode, encoding=encoding)



def open_file(path: Path, mode='r') -> TextIO:
    """
    Get file descriptor from path
    Considers whether the file is/should be gzipped
    """
    if path == '-':
        f = sys.stdin if 'r' in mode else sys.stdout
        if 'b' in mode:
            f = f.buffer
        return f
    else:
        if 't' not in mode:
            mode += 't'
        if 'w' in mode:
            path.parent.mkdir(exist_ok=True, parents=True)
        return get_open_method(path, mode, encoding='utf-8')



from vcfio.utils.str_utils import to_number


class EasyDict(dict):
    def get(self, key, default=None, empty_value='.', infer_type=False):
        """
        If single value
            If number
                return int / float
            Else
                return value
        If list
            If length = 1
                return first value
            Else
                return list
        """
        raw_val = super().get(key, default)

        if raw_val is None or raw_val == empty_value:
            return default

        if not infer_type:
            return raw_val

        if isinstance(raw_val, str) and ',' in raw_val:
            raw_val = raw_val.split(',')

        if isinstance(raw_val, list):
            if len(raw_val) == 1:
                return to_number(raw_val[0], default=raw_val)

            return [to_number(sub_val, default=sub_val) for sub_val in raw_val]

        return to_number(raw_val, default=raw_val)



import functools
from typing import AnyStr

from vcfio.utils.consts import GT_ZYGOSITY_MAP
from vcfio.utils.enums import Zygosity
from vcfio.utils.regex_patterns import GT_PATTERN


def to_number(value, default=''):
    """
    Try to convert value to int or float
    """
    try:
        x = max(int(value), float(value))
    except ValueError:
        try:
            x = float(value)
        except ValueError:
            x = default
    except TypeError:
        x = default
    return x


@functools.lru_cache(maxsize=2**16)
def standardize_chromosome(raw_chr):
    chromosome = raw_chr
    if not chromosome.startswith('chr'):
        chromosome = 'chr' + raw_chr
    if chromosome in ('chrM', 'chrMT'):
        chromosome = 'chrM'

    return chromosome


def calculate_zygosity(gt: AnyStr) -> Zygosity:
    """
    calculate zygosity of a sample
    examples:
        '0/1' -> HET
        './.' -> NO_COVERAGE
    """
    gt_pair = GT_PATTERN.findall(gt)[0]
    return GT_ZYGOSITY_MAP.get(gt_pair, Zygosity.no_coverage)


from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from vcfio.utils.file_utils import open_file
from vcfio.variant.variant import Variant

if TYPE_CHECKING:
    from typing import Union
    from typing import AnyStr
    from typing import Iterable

REQUIRED_HEADERS = (
    '##fileformat=VCFv4.2',
    '#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	{sample_names}'
)


class VcfWriter:
    def __init__(self, output_file: Union[AnyStr, Path]):
        """
        output_file -- /path/to/output
        headers -- array of vcf headers - should comply with VCF specifications
        variants -- array of Variants
        """
        self.output_file = Path(output_file)

        self._file_descriptor = open_file(self.output_file, mode='w')
        self._write_last_header = False

    def write(self, headers: Iterable[AnyStr] = (), variants: Iterable[Variant] = ()):
        """
        Write the headers and variants to self.output_path
        If headers is empty - will write REQUIRED HEADERS (using sample names from `variants` parameter
        """
        self.write_headers(headers)
        self.write_variants(variants)

    def write_headers(self, headers: Iterable[AnyStr] = ()):
        if not headers:
            headers = [REQUIRED_HEADERS[0]]
            self._write_last_header = True
        for header in headers:
            self._file_descriptor.write(header)
            self._file_descriptor.write('\n')

    def write_variants(self, variants: Iterable[Variant]):
        for variant in variants:
            self.write_variant(variant)

    def write_variant(self, variant: Variant):
        if self._write_last_header:
            self._file_descriptor.write(REQUIRED_HEADERS[-1].format(sample_names='\t'.join(variant.sample_names)))
            self._file_descriptor.write('\n')
            self._write_last_header = False

        self._file_descriptor.write(variant.to_vcf_line())
        self._file_descriptor.write('\n')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False

    def close(self):
        self._file_descriptor.close()



from __future__ import annotations

from cmath import nan
from collections import OrderedDict

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import AnyStr
    from typing import List
    from typing import Tuple

from vcfio.utils.easy_dict import EasyDict
from vcfio.utils.str_utils import standardize_chromosome
from vcfio.utils.str_utils import to_number


class VariantProperties:
    # Variant is split into 2 classes to avoid a "War and Peace" situation
    # This class is for Variant's properties and basic parsing methods

    def __init__(self, chromosome=None, position=None, vcf_id=None, ref=None, alt=None, quality=None,
                 vcf_filter=None, info=None, sample_format=None, sample_names=None, samples=None):
        """
        All parameters are raw strings, the class' properties will parse them (if needed).
        All the raw attributes are stored as _<attribute-name> and the properties are without _ .
        """
        self._chromosome = chromosome
        self._position = position
        self._vcf_id = vcf_id
        self._ref = ref
        self._alt = alt
        self._quality = quality
        if vcf_filter is None:
            vcf_filter = []
        self._vcf_filter = vcf_filter
        self._raw_info = info
        self._parsed_info = None
        self._sample_format = sample_format
        if samples is None:
            samples = []
        if sample_names is None:
            sample_names = []
        self._sample_names = sample_names
        self._raw_samples = tuple(samples)
        self._parsed_samples = None

    @property
    def chromosome(self) -> AnyStr:
        return standardize_chromosome(self._chromosome)

    @chromosome.setter
    def chromosome(self, value):
        self._chromosome = value

    @property
    def position(self):
        return int(self._position)

    @position.setter
    def position(self, value):
        self._position = value

    @property
    def vcf_id(self):
        return self._vcf_id

    @vcf_id.setter
    def vcf_id(self, value):
        self._vcf_id = value

    @property
    def ref(self):
        return self._ref

    @ref.setter
    def ref(self, value):
        self._ref = value

    @property
    def alt(self):
        if isinstance(self._alt, str):
            return self._alt.split(',')
        return self._alt

    @alt.setter
    def alt(self, value):
        self._alt = value

    @property
    def quality(self):
        return to_number(self._quality, default=0)

    @quality.setter
    def quality(self, value):
        self._quality = value

    @property
    def vcf_filter(self) -> List[AnyStr]:
        if isinstance(self._vcf_filter, str):
            return self._vcf_filter.split(';')
        if isinstance(self._vcf_filter, list):
            return self._vcf_filter
        return []

    @property
    def info(self):
        # Already parsed
        if self._parsed_info:
            return self._parsed_info

        # User set the value (example: variant.info = {'SVTYPE': 'DEL', 'DN': 'DeNovo'}
        if isinstance(self._raw_info, dict):
            self._parsed_info = EasyDict(self._raw_info)
            return self._parsed_info

        self._parsed_info = EasyDict()

        # Empty values
        if self._raw_info is None or self._raw_info == '.':
            return self._parsed_info

        # Parse
        if isinstance(self._raw_info, str):
            for pair in self._raw_info.split(';'):
                """
                key=val; --> key: val
                key=;    --> key: ''
                key;     --> key: nan
                """
                key, *val = pair.split('=', 1)
                self._parsed_info[key] = val[0] if val else nan

        return self._parsed_info

    @info.setter
    def info(self, value):
        self._raw_info = value

    @property
    def sample_format(self) -> Tuple[AnyStr]:
        if isinstance(self._sample_format, (list, tuple)):
            return tuple(self._sample_format)

        return tuple(self._sample_format.split(':'))

    @sample_format.setter
    def sample_format(self, value):
        self._sample_format = value

    @property
    def sample_names(self) -> Tuple:
        if isinstance(self._sample_names, str):
            return tuple(self._sample_names.split(','))

        return tuple(self._sample_names)

    @property
    def samples(self):
        # Already parsed
        if self._parsed_samples:
            self._parsed_samples = OrderedDict(((key, EasyDict(val)) for key, val in self._parsed_samples.items()))
            return self._parsed_samples

        # User set the values (example: variant.samples = {'DP': 10, 'GT': '1/0'})
        if isinstance(self._raw_samples, dict) and not isinstance(self._raw_samples, OrderedDict):
            self._parsed_samples = OrderedDict(((key, EasyDict(val)) for key, val in self._raw_samples.items()))
            return self._parsed_samples

        # parse the samples
        self._parsed_samples = OrderedDict()
        for name, raw_sample in zip(self.sample_names, self._raw_samples):
            self._parsed_samples[name] = EasyDict(zip(self.sample_format, raw_sample.split(':')))
        return self._parsed_samples

    @samples.setter
    def samples(self, value):
        self._raw_samples = value

    def __repr__(self):
        return f'Variant(chromosome={self.chromosome},position={self.position},ref={self.ref},alt={self.alt})'



from vcfio.utils.enums import Zygosity

VALID_CHROMOSOMES = [
    'chr1',
    'chr2',
    'chr3',
    'chr4',
    'chr5',
    'chr6',
    'chr7',
    'chr8',
    'chr9',
    'chr10',
    'chr11',
    'chr12',
    'chr13',
    'chr14',
    'chr15',
    'chr16',
    'chr17',
    'chr18',
    'chr19',
    'chr20',
    'chr21',
    'chr22',
    'chrX',
    'chrY',
    'chrM'
]

GT_ZYGOSITY_MAP = {
    ('0', '0'): Zygosity.reference,
    ('0', '1'): Zygosity.heterozygote,
    ('1', '0'): Zygosity.heterozygote,
    ('1', '1'): Zygosity.homozygote,
    ('1', '2'): Zygosity.heterozygote,
    ('2', '1'): Zygosity.heterozygote,
}